# -*- coding: utf-8 -*-
# /usr/bin/python
'''
By kyubyong park(kbpark.linguist@gmail.com) and Jongseok Kim(https://github.com/ozmig77)
https://www.github.com/kyubyong/g2p
'''
from nltk import pos_tag
from nltk.corpus import cmudict
import nltk
from nltk.tokenize import TweetTokenizer
word_tokenize = TweetTokenizer().tokenize
import numpy as np
import codecs
import re
import os
import unicodedata
from builtins import str as unicode
from .expand import normalize_numbers

try:
    nltk.data.find('taggers/averaged_perceptron_tagger.zip')
except LookupError:
    nltk.download('averaged_perceptron_tagger')
try:
    nltk.data.find('corpora/cmudict.zip')
except LookupError:
    nltk.download('cmudict')

dirname = os.path.dirname(__file__)

def construct_homograph_dictionary():
    f = os.path.join(dirname,'homographs.en')
    homograph2features = dict()
    for line in codecs.open(f, 'r', 'utf8').read().splitlines():
        if line.startswith("#"): continue # comment
        headword, pron1, pron2, pos1 = line.strip().split("|")
        homograph2features[headword.lower()] = (pron1.split(), pron2.split(), pos1)
    return homograph2features

# def segment(text):
#     '''
#     Splits text into `tokens`.
#     :param text: A string.
#     :return: A list of tokens (string).
#     '''
#     print(text)
#     text = re.sub('([.,?!]( |$))', r' \1', text)
#     print(text)
#     return text.split()

class G2p(object):
    def __init__(self):
        super().__init__()
        self.graphemes = ["<pad>", "<unk>", "</s>"] + list("abcdefghijklmnopqrstuvwxyz")
        self.phonemes = ["<pad>", "<unk>", "<s>", "</s>"] + ['AA0', 'AA1', 'AA2', 'AE0', 'AE1', 'AE2', 'AH0', 'AH1', 'AH2', 'AO0',
                                                             'AO1', 'AO2', 'AW0', 'AW1', 'AW2', 'AY0', 'AY1', 'AY2', 'B', 'CH', 'D', 'DH',
                                                             'EH0', 'EH1', 'EH2', 'ER0', 'ER1', 'ER2', 'EY0', 'EY1',
                                                             'EY2', 'F', 'G', 'HH',
                                                             'IH0', 'IH1', 'IH2', 'IY0', 'IY1', 'IY2', 'JH', 'K', 'L',
                                                             'M', 'N', 'NG', 'OW0', 'OW1',
                                                             'OW2', 'OY0', 'OY1', 'OY2', 'P', 'R', 'S', 'SH', 'T', 'TH',
                                                             'UH0', 'UH1', 'UH2', 'UW',
                                                             'UW0', 'UW1', 'UW2', 'V', 'W', 'Y', 'Z', 'ZH']
        self.g2idx = {g: idx for idx, g in enumerate(self.graphemes)}
        self.idx2g = {idx: g for idx, g in enumerate(self.graphemes)}

        self.p2idx = {p: idx for idx, p in enumerate(self.phonemes)}
        self.idx2p = {idx: p for idx, p in enumerate(self.phonemes)}

        self.vocab = cmudict.dict()
        self.load_variables()
        self.homograph2features = construct_homograph_dictionary()

    def load_variables(self):
        self.variables = np.load(os.path.join(dirname,'checkpoint20.npz'))
        self.enc_emb = self.variables["enc_emb"]  # (29, 64). (len(graphemes), emb)
        self.enc_w_ih = self.variables["enc_w_ih"]  # (3*128, 64)
        self.enc_w_hh = self.variables["enc_w_hh"]  # (3*128, 128)
        self.enc_b_ih = self.variables["enc_b_ih"]  # (3*128,)
        self.enc_b_hh = self.variables["enc_b_hh"]  # (3*128,)

        self.dec_emb = self.variables["dec_emb"]  # (74, 64). (len(phonemes), emb)
        self.dec_w_ih = self.variables["dec_w_ih"]  # (3*128, 64)
        self.dec_w_hh = self.variables["dec_w_hh"]  # (3*128, 128)
        self.dec_b_ih = self.variables["dec_b_ih"]  # (3*128,)
        self.dec_b_hh = self.variables["dec_b_hh"]  # (3*128,)
        self.fc_w = self.variables["fc_w"]  # (74, 128)
        self.fc_b = self.variables["fc_b"]  # (74,)

    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))

    def grucell(self, x, h, w_ih, w_hh, b_ih, b_hh):
        rzn_ih = np.matmul(x, w_ih.T) + b_ih
        rzn_hh = np.matmul(h, w_hh.T) + b_hh

        rz_ih, n_ih = rzn_ih[:, :rzn_ih.shape[-1] * 2 // 3], rzn_ih[:, rzn_ih.shape[-1] * 2 // 3:]
        rz_hh, n_hh = rzn_hh[:, :rzn_hh.shape[-1] * 2 // 3], rzn_hh[:, rzn_hh.shape[-1] * 2 // 3:]

        rz = self.sigmoid(rz_ih + rz_hh)
        r, z = np.split(rz, 2, -1)

        n = np.tanh(n_ih + r * n_hh)
        h = (1 - z) * n + z * h

        return h

    def gru(self, x, steps, w_ih, w_hh, b_ih, b_hh, h0=None):
        if h0 is None:
            h0 = np.zeros((x.shape[0], w_hh.shape[1]), np.float32)
        h = h0  # initial hidden state
        outputs = np.zeros((x.shape[0], steps, w_hh.shape[1]), np.float32)
        for t in range(steps):
            h = self.grucell(x[:, t, :], h, w_ih, w_hh, b_ih, b_hh)  # (b, h)
            outputs[:, t, ::] = h
        return outputs

    def encode(self, word):
        chars = list(word) + ["</s>"]
        x = [self.g2idx.get(char, self.g2idx["<unk>"]) for char in chars]
        x = np.take(self.enc_emb, np.expand_dims(x, 0), axis=0)

        return x

    def predict(self, word):
        # encoder
        enc = self.encode(word)
        enc = self.gru(enc, len(word) + 1, self.enc_w_ih, self.enc_w_hh,
                       self.enc_b_ih, self.enc_b_hh, h0=np.zeros((1, self.enc_w_hh.shape[-1]), np.float32))
        last_hidden = enc[:, -1, :]

        # decoder
        dec = np.take(self.dec_emb, [2], axis=0)  # 2: <s>
        h = last_hidden

        preds = []
        for i in range(20):
            h = self.grucell(dec, h, self.dec_w_ih, self.dec_w_hh, self.dec_b_ih, self.dec_b_hh)  # (b, h)
            logits = np.matmul(h, self.fc_w.T) + self.fc_b
            pred = logits.argmax()
            if pred == 3: break  # 3: </s>
            preds.append(pred)
            dec = np.take(self.dec_emb, [pred], axis=0)

        preds = [self.idx2p.get(idx, "<unk>") for idx in preds]
        return preds

    def lookup(self, word, pos):
        return self.vocab[word][0]

    def __call__(self, text):
        # preprocessing
        text = unicode(text)
        text = normalize_numbers(text)
        text = ''.join(char for char in unicodedata.normalize('NFD', text)
                       if unicodedata.category(char) != 'Mn')  # Strip accents
        text = text.lower()
        text = re.sub("[^ a-z'.,?!\-]", "", text)
        text = text.replace("i.e.", "that is")
        text = text.replace("e.g.", "for example")

        # tokenization
        words = word_tokenize(text)
        tokens = pos_tag(words)  # tuples of (word, tag)

        # steps
        prons = []
        for word, pos in tokens:
            if re.search("[a-z]", word) is None:
                pron = [word]

            elif word in self.homograph2features:  # Check homograph
                pron1, pron2, pos1 = self.homograph2features[word]
                if pos.startswith(pos1):
                    pron = pron1
                else:
                    pron = pron2
            elif word in self.vocab:  # lookup in vocab
                pron = self.lookup(word, pos)
            else: # predict for oov
                pron = self.predict(word)

            prons.extend(pron)
            prons.extend([" "])

        return prons[:-1]

if __name__ == '__main__':
    texts = ["I have $250 in my pocket.", # number -> spell-out
             "popular pets, e.g. cats and dogs", # e.g. -> for example
             "I refuse to collect the refuse around here.", # homograph
             "I'm an activationist."] # newly coined word
    g2p = G2p()
    for text in texts:
        out = g2p(text)
        print(out)




"""
Tokenizer module
"""

import os
import re

import numpy as np

from anyascii import anyascii

from .expand import normalize_numbers
from .g2p import G2p, dirname


class TTSTokenizer(G2p):
    """
    Text to Speech (TTS) Tokenizer - converts English Graphemes to Phoneme Tokens Ids
    """

    def __init__(self, tokens=None, nospace=True):
        """
        Creates a new tokenizer instance. Optionally can pass a list of phoneme tokens to use to map
        outputs to token id arrays.

        Args:
            tokens: list of phoneme tokens - uses order to infer phoneme token ids
            nospace: if space phoneme tokens should be removed, defaults to True
        """

        # Call parent constructor
        super().__init__()

        # Build map of phoneme token to id
        self.tokens = {token:x for x, token in enumerate(tokens)} if tokens else {}

        # Remove phoneme space tokens
        self.nospace = nospace

        # List of expansions
        self.expansions = [(re.compile(fr"\b{x[0]}\.", re.IGNORECASE), x[1]) for x in [
            ('mrs', 'misess'),
            ('mr', 'mister'),
            ('dr', 'doctor'),
            ('st', 'saint'),
            ('co', 'company'),
            ('jr', 'junior'),
            ('maj', 'major'),
            ('gen', 'general'),
            ('drs', 'doctors'),
            ('rev', 'reverend'),
            ('lt', 'lieutenant'),
            ('hon', 'honorable'),
            ('sgt', 'sergeant'),
            ('capt', 'captain'),
            ('esq', 'esquire'),
            ('ltd', 'limited'),
            ('col', 'colonel'),
            ('ft', 'fort'),
        ]]

        # Override vocab
        self.vocab = {
            **self.loadvocab(os.path.join(dirname, "cmudict.dict")),
            **self.loadvocab(os.path.join(dirname, "custom.dict"))
        }

        # Multi word terms (separated by hyphens)
        self.multiterm = {
            x.replace("-", " ").strip():x for x in
            sorted([x for x in self.vocab if "-" in x], key=lambda x: x.count("-"), reverse=True)
        }

        # Abbrevations (periods replaced with zzz)
        # Only support multiple period abbrevations)
        self.multiterm.update({
            x.replace("zzz", "."):x for x in
            sorted(
                [x for x in self.vocab if "zzz" in x[:-1]], key=lambda x: x.count("zzz"),
                reverse=True
            )
        })

    def __call__(self, text):
        # Normalize text
        text = self.normalize(text)

        # Convert to phonemes
        tokens = super().__call__(text)

        # Remove whitespace tokens
        if self.nospace:
            tokens = [x for x in tokens if x != " "]

        # Build phoneme token id array and return
        return np.array([self.tokens[x] for x in tokens], dtype=np.int64) if self.tokens else tokens

    def punctuation(self):
        """
        Gets a list of punctuation token ids.

        Returns:
            list of punctuation token ids
        """

        return [v for k, v in self.tokens.items() if k in ".,!?;"]

    def lookup(self, word, pos):
        # Primary stress for nouns, adjectives, conjunctions and personal pronouns
        primary = pos.startswith(("N", "J", "CC", "PRP"))
        return self.vocab[word][0] if primary else self.vocab[word][-1]

    def normalize(self, text):
        """
        Applies text normalization and cleaning routines for English text.

        Args:
            text: input text

        Returns:
            normalized text
        """

        # Clean and normalize text
        text = anyascii(text)
        text = self.dates(text)
        text = text.lower()
        text = normalize_numbers(text)
        text = self.expand(text)
        text = self.symbols(text)

        # Combine multi-term entries into a single term
        for k, v in self.multiterm.items():
            text = re.sub(fr"\b{re.escape(k)}\b", v, text)

        # Make uppercase to work with parent stripping accented chars logic
        text = text.upper()

        # Collapse whitespace
        text = re.sub(r"\s+", " ", text)

        return text

    def dates(self, text):
        """
        Normalizes dates.

        Args:
            text: input text

        Return:
            normalized text
        """

        text = re.sub(r"(\d+) BC", r"\1 b.c.", text)
        text = re.sub(r"(\d+) AD", r"\1 a.d.", text)

        return text

    def expand(self, text):
        """
        Runs a set of text expansions.

        Args:
            text: input text

        Returns:
            expanded text
        """

        for regex, replacement in self.expansions:
            text = re.sub(regex, replacement, text)

        return text

    def symbols(self, text):
        """
        Expands and cleans symbols from text.

        Args:
            text: input text

        Returns:
            clean text
        """

        # Expand symbols
        text = re.sub(r"\;", ",", text)
        text = re.sub(r"\:", ",", text)
        text = re.sub(r"\-", " ", text)
        text = re.sub(r"\&", "and", text)

        # Clean unnecessary symbols
        return re.sub(r'[\(\)\[\]\<\>\"]+', '', text)

    def loadvocab(self, path):
        """
        Loads a vocabulary file from path. This file has a word to pronunciation mapping.
        Pronunciations use the ARPABET codes.

        Args:
            path: path to vocab file

        Returns:
            {word: [pronunciations]}
        """

        vocab = {}

        with open(path, encoding="utf-8") as f:
            for line in f:
                if line.strip() and not line.startswith("#"):
                    word, pron = line.strip().split(maxsplit=1)

                    # Normalize word
                    word = re.sub(r"\(\d+\)", r"", word).lower().strip()

                    # Create special entries for multi-period abbrevations
                    if "." in word and (not word.endswith(".") or word.count(".") > 1):
                        word = word.replace(".", "zzz")

                    # Remove trailing comments, and split into tokens
                    pron = re.sub(r"(.+?)#.*", r"\1", pron).strip().split()

                    # Initialize word
                    if word not in vocab:
                        vocab[word] = []

                    # Add pronunciation
                    vocab[word].append(pron)

        return vocab


class IPATokenizer(TTSTokenizer):
    """
    International Phonetic Alphabet (IPA) TTS Tokenizer - converts English Graphemes to
    IPA Tokens Ids
    """

    def __init__(self, tokenize=True, transcribe=True):
        """
        Creates a new IPA tokenizer instance.

        Args:
            tokenize: if True, IPA token ids are returned (default) otherwise, the IPA text is
                      returned
            transcribe: if True, input text is transcribed to IPA (default), otherwise it is assumed
                        that the text is already transcribed to IPA
        """

        # Call parent constructor. This tokenizer has it's own token mapping and requires spaces
        super().__init__(None, False)

        # Enables tokenization to IPA token ids
        self.tokenize = tokenize

        # Enables text to IPA transcription
        self.transcribe = transcribe

    def __call__(self, text):
        if self.transcribe:
            # Tokenize text using base tokenization method
            tokens = super().__call__(text)

            # Get ARPABET to IPA vocab mapping
            vocab = self.ipavocab()

            # Map ARPABET tokens to IPA tokens and join as a string
            text = "".join(vocab.get(x, x) for x in tokens)
        else:
            text = text.strip()

        # IPA token lookup
        tokens = self.ipatokens()

        # Lookup token ids and return
        return np.array([tokens.get(x, x) for x in text], dtype=np.int64) if self.tokenize else text

    def punctuation(self):
        """
        Gets a list of punctuation token ids.

        Returns:
            list of punctuation token ids
        """

        return [v for k, v in self.ipatokens().items() if k in ".,!?;"]

    def ipavocab(self):
        """
        Builds a vocabulary mapping the ARPABET alphabet (via CMUdict) to the
        International Phonetic Alphabet (IPA).

        See the following links for more information on this mapping.
          - https://github.com/Kyubyong/g2p/issues/29
          - https://en.wikipedia.org/wiki/ARPABET

        Returns:
            ARPABET to IPA vocabulary mapping
        """

        return {
            '<pad>': '<pad>',
            '<unk>': '<unk>',
            '<s>': '<s>',
            '</s>': '</s>',
            'AA0': 'ɑ',
            'AA1': 'ˈɑː',
            'AA2': 'ˌɑ',
            'AE0': 'æ',
            'AE1': 'ˈæ',
            'AE2': 'ˌæ',
            'AH0': 'ə',
            'AH1': 'ˈʌ',
            'AH2': 'ˌʌ',
            'AO0': 'ɔ',
            'AO1': 'ˈɔː',
            'AO2': 'ˌɔ',
            'AW0': 'aʊ',
            'AW1': 'ˈaʊ',
            'AW2': 'ˌaʊ',
            'AY0': 'aɪ',
            'AY1': 'ˈaɪ',
            'AY2': 'ˌaɪ',
            'B': 'b',
            'CH': 'tʃ',
            'D': 'd',
            'DH': 'ð',
            'EH0': 'ɛ',
            'EH1': 'ˈɛ',
            'EH2': 'ˌɛ',
            'ER0': 'ɚ',
            'ER1': 'ˈɚ',
            'ER2': 'ˌɚ',
            'EY0': 'eɪ',
            'EY1': 'ˈeɪ',
            'EY2': 'ˌeɪ',
            'F': 'f',
            'G': 'ɡ',
            'HH': 'h',
            'IH0': 'ɪ',
            'IH1': 'ˈɪ',
            'IH2': 'ˌɪ',
            'IY0': 'i',
            'IY1': 'ˈiː',
            'IY2': 'ˌi',
            'JH': 'dʒ',
            'K': 'k',
            'L': 'l',
            'M': 'm',
            'N': 'n',
            'NG': 'ŋ',
            'OW0': 'oʊ',
            'OW1': 'ˈoʊ',
            'OW2': 'ˌoʊ',
            'OY0': 'ɔɪ',
            'OY1': 'ˈɔɪ',
            'OY2': 'ˌɔɪ',
            'P': 'p',
            'R': 'ɹ',
            'S': 's',
            'SH': 'ʃ',
            'T': 't',
            'TH': 'θ',
            'UH0': 'ʊ',
            'UH1': 'ˈʊ',
            'UH2': 'ˌʊ',
            'UW0': 'u',
            'UW1': 'ˈuː',
            'UW2': 'ˌu',
            'V': 'v',
            'W': 'w',
            'Y': 'j',
            'Z': 'z',
            'ZH': 'ʒ',
        }

    def ipatokens(self):
        """
        Builds an International Phonetic Alphabet (IPA) token to token id mapping.

        This grapheme mapping is from here: https://hf.co/hexgrad/Kokoro-82M/blob/main/kokoro.py#L74

        Returns:
            {token: token id}
        """

        # pylint: disable=C0301
        pad = "$"
        punctuation = ';:,.!?¡¿—…"«»“” '
        letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
        ipaletters = "ɑɐɒæɓʙβɔɕçɗɖðʤəɘɚɛɜɝɞɟʄɡɠɢʛɦɧħɥʜɨɪʝɭɬɫɮʟɱɯɰŋɳɲɴøɵɸθœɶʘɹɺɾɻʀʁɽʂʃʈʧʉʊʋⱱʌɣɤʍχʎʏʑʐʒʔʡʕʢǀǁǂǃˈˌːˑʼʴʰʱʲʷˠˤ˞↓↑→↗↘'̩'ᵻ"

        symbols = [pad] + list(punctuation) + list(letters) + list(ipaletters)
        mapping = {}
        for x, token in enumerate(symbols):
            mapping[token] = x

        return mapping



# pylint: disable = C0111
from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    DESCRIPTION = f.read()

setup(
    name = "ttstokenizer",
    version = "1.1.0",
    description = "Tokenizer for Text to Speech (TTS) models",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    url = "https://github.com/neuml/ttstokenizer",
    license="Apache 2.0: http://www.apache.org/licenses/LICENSE-2.0",
    packages = ["ttstokenizer"],
    install_requires = [
        "anyascii>=0.3.1",
        "inflect>=0.3.1",
        "numpy>=1.13.1",
        "nltk>=3.2.4"
    ],
    include_package_data = True
)
